/*-------------------------------------------------------------------------+
| Copyright 2020 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view.fx;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.base.utils.AnnotationUtils.getAnnotationValueProvider;
import static org.fortiss.tooling.kernel.ui.util.SelectionUtils.checkAndPickFirst;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.getChildrenWithType;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.emf.common.util.EList;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.annotation.AnnotationValueService;
import org.fortiss.tooling.base.annotation.IAnnotationValueService;
import org.fortiss.tooling.base.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.view.generic.ColumnHandle;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTreeContentProviderBase;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTreeTableUIProviderBase;
import org.fortiss.tooling.common.ui.javafx.control.treetableview.DynamicTreeTableViewer;
import org.fortiss.tooling.common.ui.javafx.layout.CompositeFXControllerBase;
import org.fortiss.tooling.kernel.model.INamedCommentedElement;

import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.SplitPane;
import javafx.scene.control.TreeTableView;

/**
 * 
 * @author bayha
 */
@SuppressWarnings("unchecked")
public class AnnotationViewFXController extends CompositeFXControllerBase<SplitPane, Node>
		implements ISelectionListener {

	/**  */
	@FXML
	private SplitPane annotationSplitPane;

	/**  */
	@FXML
	private TreeTableView<AnnotationEntry> annotationTreeTableView;

	/**  */
	private DynamicTreeTableViewer<AnnotationEntry> annotationTreeViewer;

	/**  */
	private IModelElement selectedElement;

	/**  */
	private DynamicTreeContentProviderBase<AnnotationEntry> contentProvider =
			new AnnotationTableContentProvider();

	/**  */
	private DynamicTreeTableUIProviderBase<AnnotationEntry> uiProvider =
			new AnnotationTreeTableUIProvider();

	/**
	 * Associates the handles of the annotations with their ordered column indexes. Used for the UI
	 * providers of cell items.
	 */
	Map<Integer, ColumnHandle<IAnnotatedSpecification>> colIdxAnnotationMap = new HashMap<>();

	/** {@inheritDoc} */
	@Override
	public String getFXMLLocation() {
		return "AnnotationViewFx.fxml";
	}

	/** {@inheritDoc} */
	@Override
	public void initialize() {
		// EMPTY
	}

	/** {@inheritDoc} */
	@Override
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		selectedElement = checkAndPickFirst(selection, IModelElement.class);

		updateAnnotationTree();
	}

	/**
	 * 
	 */
	private void updateAnnotationTree() {
		if(selectedElement == null) {
			return;
		}

		EList<IModelElement> childElements =
				getChildrenWithType(selectedElement, IModelElement.class);

		ArtificialRoot root = new ArtificialRoot(selectedElement);

		for(IModelElement elem : childElements) {
			AnnotationEntry entry = IAnnotationValueService.getInstance().getAnnotationEntry(elem);

			if(entry != null) {
				root.elements.add(entry);
			}
		}

		// Sorted set of {@link ColumnHandle} used to instantiate this {@link
		// GenericAnnotationView}'s columns in the right order.
		Set<ColumnHandle<IAnnotatedSpecification>> sortedColumnHandles = new TreeSet<>();
		Set<AnnotationEntry> annotationEntries = new HashSet<>();
		childElements.forEach(e -> annotationEntries
				.add(AnnotationValueService.getInstance().getAnnotationEntry(e)));

		// Aggregate required columns. Column order is defined by ColumnHandle.compareTo().
		for(AnnotationEntry entry : annotationEntries) {
			for(IAnnotatedSpecification spec : entry.getSpecificationsList()) {
				ColumnHandle<IAnnotatedSpecification> columnHandle =
						new ColumnHandle<IAnnotatedSpecification>(entry, spec);
				sortedColumnHandles.add(columnHandle);
			}
		}

		annotationTreeTableView.getColumns().clear();

		boolean showRoot = false;
		int revealLevel = 1;
		annotationTreeViewer = new DynamicTreeTableViewer<AnnotationEntry>(annotationTreeTableView,
				root, showRoot, revealLevel, contentProvider, uiProvider);

		annotationTreeViewer.addColumn("Model Element", 200);
		annotationTreeViewer.addColumn("Comment", 250);

		int columnIdx = 2;
		for(ColumnHandle<IAnnotatedSpecification> handle : sortedColumnHandles) {
			annotationTreeViewer.addColumn(handle.getColumnName(), 150);
			colIdxAnnotationMap.put(columnIdx++, handle);
		}
	}

	/**
	 * 
	 * @author bayha
	 */
	protected class AnnotationTreeTableUIProvider
			extends DynamicTreeTableUIProviderBase<AnnotationEntry> {

		/** {@inheritDoc} */
		@Override
		public String getLabel(AnnotationEntry entry, int column) {
			IModelElement element = entry.getModelElement();

			switch(column) {
				case 0:
					if(element instanceof INamedCommentedElement) {
						return ((INamedCommentedElement)element).getName();
					}
					return "";
				case 1:
					if(element instanceof INamedCommentedElement) {
						return ((INamedCommentedElement)element).getComment();
					}
					return "";
				default:
					ColumnHandle<IAnnotatedSpecification> colHandle =
							colIdxAnnotationMap.get(column);
					IAnnotationValueProvider<IAnnotatedSpecification> valProvider =
							getAnnotationValueProvider(element,
									colHandle.getAnnotatedSpecificationType());
					try {
						return valProvider.getAnnotationValue(colHandle.getAnnotatedSpecification())
								.toString();
					} catch(Exception e) {
						// For this demo class, it is expected that direct conversions to
						// strings will fail for some annotations. Simple don't show anything in
						// this case.
						return "";
					}
			}
		}
	}

	/**
	 * 
	 * @author bayha
	 */
	protected class AnnotationTableContentProvider
			extends DynamicTreeContentProviderBase<AnnotationEntry> {

		/** {@inheritDoc} */
		@Override
		protected Collection<? extends AnnotationEntry> getChildren(AnnotationEntry parent) {
			if(parent instanceof ArtificialRoot) {
				return ((ArtificialRoot)parent).elements;
			}

			return emptyList();
		}
	}

	/**
	 * 
	 * @author bayha
	 */
	protected class ArtificialRoot extends AnnotationEntry {

		/**
		 * @param modelElement
		 */
		public ArtificialRoot(IModelElement modelElement) {
			super(modelElement);
		}

		/**  */
		List<AnnotationEntry> elements = new ArrayList<AnnotationEntry>();
	}
}
